"""Internal module for shared utilities within sorunlib.

The usual caveats apply, these interfaces might change without notice.

"""

import datetime as dt
import time

import ocs

from sorunlib.commands import _timestamp_to_utc_datetime


def _check_error(client, response):
    """Check if a response is an error or timeout."""
    op = response.session['op_name']
    instance = client.instance_id

    if response.status == ocs.ERROR:
        error = f"Request for Operation {op} in Agent {instance} failed.\n" + \
            str(response)
        raise RuntimeError(error)
    elif response.status == ocs.TIMEOUT:
        error = f"Timeout reached waiting for {op} in Agent {instance} to " + \
            "complete.\n" + str(response)
        raise RuntimeError(error)


def check_response(client, response):
    """Check that a response from OCS indicates successful task/process
    completion.

    Args:
        client (ocs.ocs_client.OCSClient): OCS Client which returned the
            response.
        response (ocs.ocs_client.OCSReply): Response from an OCS operation
            call.

    Raises:
        RuntimeError: When Operation has not completed successfully (either on
            initial request or after completion) or the response has timed out.

    """
    op = response.session['op_name']
    instance = client.instance_id

    _check_error(client, response)

    if not response.session['success']:
        error = f'Task {op} in Agent {instance} failed to complete " + \
            "successfully.\n' + str(response)
        raise RuntimeError(error)


def check_running(client, response):
    """Check that a response from OCS indicates a task/process is still
    running.

    Args:
        client (ocs.ocs_client.OCSClient): OCS Client which returned the
            response.
        response (ocs.ocs_client.OCSReply): Response from an OCS operation
            call.

    Raises:
        RuntimeError: When Operation is not running.

    """
    op = response.session['op_name']
    instance = client.instance_id

    _check_error(client, response)
    if response.session['status'] != 'running':
        error = f"Operation {op} in Agent {instance} is not in the 'running' " + \
            "state.\n" + str(response)
        raise RuntimeError(error)


def check_started(client, response, timeout=60):
    """Check that a process has started and is in the 'running' state. If it
    has not finished starting, wait for timeout seconds for it to start. If
    it is not running by then, raise exception.

    This is meant to be called right after a process' start call is made, to
    ensure that process starts successfully. Use ``check_running()`` if you
    want to check that an already running process is still running.

    Args:
        client (ocs.ocs_client.OCSClient): OCS Client which returned the
            response.
        response (ocs.ocs_client.OCSReply): Response from an OCS operation
            call.
        timeout (float): How long to wait, in seconds, for the operation to go
            from 'starting' to 'running' before raising an exception.

    Raises:
        RuntimeError: When Operation does not properly transition from
            'starting' to 'running' or is not 'running'.

    """
    op = response.session['op_name']
    instance = client.instance_id

    _check_error(client, response)

    op_code = response.session.get('op_code')

    if op_code == 2:  # STARTING
        _operation = client.__getattribute__(op)

        # Wait at most ~timeout seconds while checking the status
        for i in range(timeout):
            response = _operation.status()
            op_code = response.session.get('op_code')
            if op_code == 3:
                # Tricky to change state during testing w/little reward
                return  # pragma: no cover
            time.sleep(1)

        error = f"Check timed out. Operation {op} in Agent {instance} stuck in " + \
            "'starting' state.\n" + str(response)
        raise RuntimeError(error)

    if op_code != 3:  # RUNNING
        error = f"Operation {op} in Agent {instance} is not 'running'.\n" + \
            f"Current OpCode: {op_code}\n" + str(response)
        raise RuntimeError(error)


def _check_operation_running(client, operation):
    op = client.__getattribute__(operation)
    resp = op.status()
    check_running(client, resp)


def _seconds_until_target(target):
    target_dt = _timestamp_to_utc_datetime(target)
    now = dt.datetime.now(dt.timezone.utc)
    diff = (target_dt - now).total_seconds()
    return max(diff, 0)


def monitor_process(client, operation, stop_time, check_interval=10):
    """Monitor a running process.

    This will block until ``stop_time``, checking that the process given in
    ``operation`` is still running every ``check_interval`` seconds.

    Args:
        client (ocs.ocs_client.OCSClient): OCS Client that has the process to
            monitor.
        operation (str): Operation name to monitor.
        stop_time (str): Time in ISO format and in UTC timezone to stop
            monitoring the process. If UTC ("+00:00") is not explicitly used in
            the timestamp it is assumed.
        check_interval (float): Interval, in seconds, at which the Operation
            status is checked. Defaults to 10 seconds.

    Raises:
        RuntimeError: If Operation stops running before ``stop_time``.

    """
    diff = _seconds_until_target(stop_time)

    while diff > 0:
        # Check process response
        _check_operation_running(client, operation)

        # Wait until next check
        if diff > check_interval:
            time.sleep(check_interval)
        else:
            # Recompute diff to avoid waiting too long
            diff = _seconds_until_target(stop_time)
            time.sleep(diff)

        # Recompute diff
        diff = _seconds_until_target(stop_time)



import os
os.environ["OCS_CONFIG_DIR"] = "./test_util/"
os.environ["SORUNLIB_CONFIG"] = "./data/example_config.yaml"

from sorunlib.config import load_config

CONFIG = {'smurf_failure_threshold': 2,
          'registry': 'registry',
          'wiregrid': {'labjack': 'wg-labjack'}}


def test_load_config():
    cfg = load_config("./data/example_config.yaml")
    assert cfg == CONFIG


def test_load_config_from_env():
    cfg = load_config()
    assert cfg == CONFIG



import time
import datetime as dt


def _timestamp_to_utc_datetime(timestamp):
    """Produce a UTC datetime object from a naive or UTC timestamp.

    Args:
        timestamp (str): Time in ISO format and in UTC timezone,
            i.e. "2015-10-21T07:28:00", "2023-01-01T0:00:00+00:00"

    Returns:
        datetime.datetime: datetime object with datetime.timezone.utc tzinfo.

    Raises:
        ValueError: If `timestamp` has an unsupported timezone.

    """
    target = dt.datetime.fromisoformat(timestamp)

    # Effectively pytz.utc.localize(target)
    if target.tzinfo is None:
        target = target.replace(tzinfo=dt.timezone.utc)

    # Check timezone
    if target.tzinfo is not dt.timezone.utc:
        offset = target.tzinfo.tzname(None)
        raise ValueError(f'Unsupported timezone ({offset}) detected. '
                         + 'Timezone must be UTC.')

    return target


def wait_until(timestamp, tolerance=None):
    """Wait until a specified time.

    Args:
        timestamp (str): Time in ISO format and in UTC timezone to wait
            until. If UTC ("+00:00") is not explicitly used in the timestamp it
            is assumed.
        tolerance (int, float, or str): Tolerance on the difference between the
            time when the function is evaluated and `timestamp`. Can be
            specified as a ISO formatted timestamp (str) or in number of
            seconds (int, float).

            When evaluated, if `tolerance` (str) is in the past or if the
            difference between the current time and `timestamp` is greater than
            this `tolerance` (int, float), then an error is raised. If `None`,
            an error will not be raised. Default is `None`.

    Raises:
        ValueError: If `timestamp` has an unsupported timezone, `tolerance`
            is an unsupported type, or if the current time at evaluation is past
            the threshold set by the `tolerance`.

    Examples:
        >>> wait_until("2015-10-21T07:28:00")
        >>> wait_until("2015-10-21T07:28:00+00:00")
        >>> wait_until("2015-10-21T07:28:00+00:00", 60)
        >>> wait_until("2015-10-21T07:28:00+00:00", "2015-10-21T07:29:00+00:00)

    """
    target = _timestamp_to_utc_datetime(timestamp)

    # Grab current UTC timestamp
    now = dt.datetime.now(dt.timezone.utc)

    # Determine "deadline" from tolerance
    if tolerance is None:
        deadline = None
    elif isinstance(tolerance, (int, float)):
        deadline = target + dt.timedelta(seconds=tolerance)
    elif isinstance(tolerance, str):
        deadline = _timestamp_to_utc_datetime(tolerance)
    else:
        raise ValueError(f"Unsupported type {type(tolerance)} provide for tolerance {tolerance}")

    # Raise error if currently past the "deadline"
    if deadline is None:
        pass
    elif now > deadline:
        raise ValueError(f"Current time ({timestamp}) is past deadline "
                         + f"({deadline.isoformat()}) set by tolerance ({tolerance})")

    # Wait until timestamp
    if target > now:
        duration = (target - now).total_seconds()
        print(f"Waiting for {duration} seconds")
        time.sleep(duration)
    else:
        diff = (now - target).total_seconds()
        print(f"No wait, as target is {diff} seconds in the past")



import time

import sorunlib as run
from sorunlib._internal import check_response, check_running

EL_DIFF_THRESHOLD = 0.5  # deg diff from 50 that its ok to run calibration
BORESIGHT_DIFF_THRESHOLD = 0.5  # deg
AGENT_TIMEDIFF_THRESHOLD = 5  # sec
OP_TIMEOUT = 60


# Internal Helper Functions
def _check_process_data(process, last_timestamp):
    """Check the latest timestamp from a process' session.data is recent enough.

    Args:
        process (str): Descriptive name, used in the error message.
        last_timestamp (float): Latest timestamp from the session.data.

    Raises:
        RuntimeError: When the last timestamp was more than
            AGENT_TIMEDIFF_THRESHOLD old.

    """
    try:
        assert ((time.time() - last_timestamp) < AGENT_TIMEDIFF_THRESHOLD)
    except AssertionError:
        error = f"{process} has no updated data. Cannot proceed with " + \
            "wiregrid calibration. Aborting."
        raise RuntimeError(error)


def _verify_temp_response(response, sensor, min_temp):
    """Check temperature above minimum temperation for safe operation of wiregrid.

    Args:
        response (ocs.ocs_client.OCSReply): Client response from the labjack
            acq process.
        sensor (str): Sensor name in the labjack session data, i.e. 'AIN0'.
        min_temp (int, float): Minimum temperature for safe operation in C.

    Raises:
        RuntimeError: When the temperature of sensor below ``min_temp``.

    """
    temp = response.session['data']['data'][sensor]
    try:
        assert (temp > min_temp)
    except AssertionError:
        error = f"Sensor {sensor} reads {temp} C, which is below minimum " + \
            f"temperature of {min_temp} C. Cannot proceed with wiregrid " + \
            "calibration. Aborting"
        raise RuntimeError(error)


# Calibration Functions
def _check_telescope_position():
    # Get current telescope position
    acu = run.CLIENTS['acu']
    resp = acu.monitor.status()
    az = resp.session['data']['StatusDetailed']['Azimuth current position']
    el = resp.session['data']['StatusDetailed']['Elevation current position']
    boresight = resp.session['data']['StatusDetailed']['Boresight current position']

    # Check appropriate elevation
    try:
        assert (abs(el - 50) < EL_DIFF_THRESHOLD)
    except AssertionError:
        error = "Telescope not at 50 deg elevation. Cannot proceed with " + \
                f"wiregrid calibration in current position ({az}, {el}). " + \
                "Aborting."
        raise RuntimeError(error)

    # Check boresight angle
    try:
        assert (abs(boresight - 0) < BORESIGHT_DIFF_THRESHOLD)
    except AssertionError:
        error = "Telescope not at 0 deg boresight. Cannot proceed with " + \
                f"wiregrid calibration in current position ({boresight}). " + \
                "Aborting."
        raise RuntimeError(error)


def _configure_power(continuous):
    """Configure the KIKUSUI power supply for rotation based on rotation type.

    Args:
        continuous (bool): Configure for continuous rotation or not.

    """
    kikusui = run.CLIENTS['wiregrid']['kikusui']

    resp = kikusui.set_v(volt=12)
    check_response(kikusui, resp)

    if continuous:
        current = 3.0
    else:
        current = 2.4

    resp = kikusui.set_c(current=current)
    check_response(kikusui, resp)


def _rotate_continuously(duration):
    kikusui = run.CLIENTS['wiregrid']['kikusui']

    # Start rotation
    resp = kikusui.set_on()
    check_response(kikusui, resp)

    # Wait
    time.sleep(duration)

    # Stop rotation
    resp = kikusui.set_off()
    check_response(kikusui, resp)


def _check_motor_on():
    actuator = run.CLIENTS['wiregrid']['actuator']

    # Check motor is on
    resp = actuator.acq.status()
    if resp.session['data']['fields']['motor'] == 1:
        print("Wiregrid motor already on.")
    # Turn on motor if needed
    elif resp.session['data']['fields']['motor'] == 0:
        resp = actuator.motor_on()
        check_response(actuator, resp)
    else:
        raise RuntimeError("Motor state unknown. Aborting...")


def _check_agents_online():
    """Check OCS agents related to the wiregrid are running."""
    actuator = run.CLIENTS['wiregrid']['actuator']
    kikusui = run.CLIENTS['wiregrid']['kikusui']
    encoder = run.CLIENTS['wiregrid']['encoder']
    labjack = run.CLIENTS['wiregrid']['labjack']

    # wiregrid_actuator
    resp = actuator.acq.status()
    last_timestamp = resp.session['data']['timestamp']
    check_running(actuator, resp)
    _check_process_data("Actuator agent", last_timestamp)

    # wiregrid_kikusui
    resp = kikusui.IV_acq.status()
    last_timestamp = resp.session['data']['timestamp']
    check_running(kikusui, resp)
    _check_process_data("Kikusui agent", last_timestamp)

    # wiregrid_encoder
    resp = encoder.acq.status()
    last_timestamp = resp.session['data']['timestamp']
    check_running(encoder, resp)
    _check_process_data("Encoder agent", last_timestamp)

    # labjack
    resp = labjack.acq.status()
    last_timestamp = resp.session['data']['timestamp']
    check_running(labjack, resp)
    _check_process_data("Labjack agent", last_timestamp)

    # encoder data stream
    resp = encoder.acq.status()
    last_timestamp = resp.session['data']['fields']['encoder_data']['last_updated']
    _check_process_data("Encoder BBB", last_timestamp)


def _check_temperature_sensors():
    """Check temperatures within safe range."""
    labjack = run.CLIENTS['wiregrid']['labjack']
    resp = labjack.acq.status()
    # AIN0 > -10 C && AIN1 > -10 C (On the wiregrid)
    _verify_temp_response(resp, 'AIN0C', -10)
    _verify_temp_response(resp, 'AIN1C', -10)
    # AIN2 > 0 C (Inside the electronics enclosure for the gridloader)
    _verify_temp_response(resp, 'AIN2C', 0)


# Public API
def insert():
    """Insert the wiregrid."""
    actuator = run.CLIENTS['wiregrid']['actuator']
    resp = actuator.insert()
    try:
        check_response(actuator, resp)
    except RuntimeError as e:
        error = "Wiregrid insertion failed. Please inspect wiregrid before " + \
                "continuing observations.\n" + str(e)
        raise RuntimeError(error)


def eject():
    """Eject the wiregrid."""
    actuator = run.CLIENTS['wiregrid']['actuator']
    resp = actuator.eject()
    try:
        check_response(actuator, resp)
    except RuntimeError as e:
        error = "Wiregrid eject failed. Please inspect wiregrid before " + \
                "continuing observations.\n" + str(e)
        raise RuntimeError(error)


def rotate(continuous, duration=30, num_laps=1, stopped_time=10.):
    """Rotate the wiregrid.

    Rotation is either continuously for ``duration`` seconds or stepwise for
    ``num_laps``, stopping every 22.5 degrees for ``stopped_time`` seconds.

    This function sets the Kikusui power settings appropriately based on the
    rotation type.

    Args:
        continuous (bool): Rotate the grid continuously if True or not if
            False.
        duration (int, float): Amount of time in seconds to rotate if rotating
            continuously. Defaults to 30 seconds.
        num_laps (int): Number of revolutions if rotating stepwise (i.e.
            ``continuous = False``). Defaults to 1.
        stopped_time (float): Duration of each 22.5 deg step in seconds if
            rotating stepwise. Defaults to 10 seconds.

    """
    kikusui = run.CLIENTS['wiregrid']['kikusui']

    # Configure power settings based on rotation type
    _configure_power(continuous)

    if continuous:
        _rotate_continuously(duration)
    else:
        resp = kikusui.stepwise_rotation(num_laps=num_laps,
                                         stopped_time=stopped_time)
        check_response(kikusui, resp)


def calibrate(continuous=False):
    """Run a wiregrid calibration.

    Args:
        continuous (bool): Calibration by continuous rotation or not.
            Default is False, in which the wiregrid rotates step-wisely.

    """
    try:
        _check_telescope_position()
        _check_agents_online()
        _check_temperature_sensors()
        _check_motor_on()

        # Rotate for reference before insertion
        rotate(continuous=True, duration=5)

        # Enable SMuRF streams
        if continuous:
            rotation = 'wg_continuous'
        else:
            rotation = 'wg_stepwise'
        run.smurf.stream('on', tag=f'wiregrid, {rotation}', subtype='cal')

        # Insert the wiregrid
        insert()

        # Rotate the wiregrid
        rotate(continuous)

        # Eject the wiregrid
        eject()
    finally:
        # Stop SMuRF streams
        run.smurf.stream('off')



import os
os.environ["OCS_CONFIG_DIR"] = "./test_util/"

import pytest

from ocs.ocs_client import OCSReply
from sorunlib import acu

from util import create_patch_clients


patch_clients_satp = create_patch_clients('satp')
patch_clients_lat = create_patch_clients('ccat')


def test_move_to(patch_clients_satp):
    acu.move_to(180, 60)
    acu.run.CLIENTS['acu'].go_to.assert_called_with(az=180, el=60)


def test_move_to_failed(patch_clients_satp):
    mocked_response = OCSReply(
        0, 'msg', {'success': False, 'op_name': 'go_to'})
    acu.run.CLIENTS['acu'].go_to.side_effect = [mocked_response]
    with pytest.raises(RuntimeError):
        acu.move_to(180, 90)


def test_set_boresight(patch_clients_satp):
    acu.set_boresight(20)
    acu.run.CLIENTS['acu'].set_boresight.assert_called_with(target=20)


def test_set_boresight_lat(patch_clients_lat):
    with pytest.raises(RuntimeError):
        acu.set_boresight(20)


def test_set_scan_params(patch_clients_satp):
    acu.set_scan_params(az_speed=2, az_accel=2, reset=True)
    acu.run.CLIENTS['acu'].set_scan_params.assert_called_with(
        az_speed=2,
        az_accel=2,
        reset=True)



import sorunlib as run
from sorunlib._internal import check_response


# Public API
def set_freq(freq):
    """Set the rotational frequency of the HWP.

    Args:
        freq (float): Target frequency to rotate the HWP in Hz. This is a
            *signed float*, the meaning of which depends on the OCS site
            configuration. For details see the `documentation for the HWP
            Supervisor Agent <docs_>`_.

    .. _docs: https://socs.readthedocs.io/en/main/agents/hwp_supervisor_agent.html

    """
    hwp = run.CLIENTS['hwp']
    resp = hwp.pid_to_freq(target_freq=freq)
    check_response(hwp, resp)


def stop(active=True, brake_voltage=None):
    """Stop the HWP.

    Args:
        active (bool, optional): If True, actively try to stop the HWP by
            applying the brake. If False, simply turn off the PMX power and let
            it spin down on its own. Defaults to True.
        brake_voltage (float, optional): Voltage used when actively stopping
            the HWP. Only considered when active is True.

    """
    hwp = run.CLIENTS['hwp']

    if active:
        if brake_voltage is None:
            resp = hwp.brake()
        else:
            resp = hwp.brake(brake_voltage=brake_voltage)
        check_response(hwp, resp)
    else:
        resp = hwp.pmx_off()
        check_response(hwp, resp)



from unittest.mock import MagicMock, patch

from util import mocked_clients

import sorunlib
from sorunlib import initialize


@patch('sorunlib.commands.time.sleep', MagicMock())
@patch('sorunlib.create_clients', mocked_clients)
def test_initialize():
    initialize()
    assert sorunlib.CLIENTS is not None



import sorunlib as run
from sorunlib._internal import check_response, check_started, monitor_process


OP_TIMEOUT = 60


def scan(description, stop_time, width, az_drift=0, tag=None, subtype=None):
    """Run a constant elevation scan, collecting detector data.

    Args:
        description (str): Description of the field/object being scanned.
        stop_time (str): Time in ISO format to scan until, i.e.
            "2022-06-21T15:58:00"
        width (float): Scan width in azimuth. The scan will start at the
            current position and move in the positive azimuth direction.
        az_drift (float): Drift velocity in deg/s, causing scan extrema to move
            accordingly.
        tag (str, optional): Tag or comma-separated listed of tags to attach to
            the operation. Passed through to the smurf stream command.
        subtype (str, optional): Operation subtype used to tag the stream.

    """
    acu = run.CLIENTS['acu']

    # Enable SMuRF streams
    run.smurf.stream('on', subtype=subtype, tag=tag)

    try:
        # Grab current telescope position
        resp = acu.monitor.status()
        az = resp.session['data']['StatusDetailed']['Azimuth current position']
        el = resp.session['data']['StatusDetailed']['Elevation current position']

        # Start telescope motion
        # az_speed and az_accel assumed from ACU defaults
        # Can be modified by acu.set_scan_params()
        resp = acu.generate_scan.start(az_endpoint1=az,
                                       az_endpoint2=az + width,
                                       el_endpoint1=el,
                                       el_endpoint2=el,
                                       el_speed=0,
                                       az_drift=az_drift)
        check_started(acu, resp)

        # Wait until stop time
        monitor_process(acu, 'generate_scan', stop_time)
    finally:
        print("Stopping scan.")
        # Stop SMuRF streams
        run.smurf.stream('off')

        # Stop motion
        acu.generate_scan.stop()
        resp = acu.generate_scan.wait(timeout=OP_TIMEOUT)
        check_response(acu, resp)
        print("Scan finished.")



import sorunlib as run
from sorunlib._internal import check_response


def move_to(az, el):
    """Move telescope to specified coordinates.

    Args:
        az (float): destination angle for the azimuthal axis
        el (float): destination angle for the elevation axis

    """
    acu = run.CLIENTS['acu']
    resp = acu.go_to(az=az, el=el)
    check_response(acu, resp)


def set_boresight(target):
    """Move the third axis to a specific target angle.

    Args:
        target (float): destination angle for boresight rotation

    Raises:
        RuntimeError: If boresight is passed to a non-satp platform.

    """
    acu = run.CLIENTS['acu']

    # Check platform type
    resp = acu.monitor.status()
    platform = resp.session['data']['PlatformType']
    if platform == "satp":
        resp = acu.set_boresight(target=target)
        check_response(acu, resp)
    else:
        raise RuntimeError(f"Platform type {platform} does not support boresight motion")


def set_scan_params(az_speed, az_accel, reset=False):
    """Update the default scan parameters, used during :func:`sorunlib.seq.scan`.

    Args:
        az_speed (float, optional): The azimuth scan speed.
        az_accel (float, optional): The (average) azimuth acceleration at
            turn-around.
        reset (bool, optional): If True, reset all params to default values
            before applying any updates passed explicitly here.

    """
    acu = run.CLIENTS['acu']
    resp = acu.set_scan_params(az_speed=az_speed,
                               az_accel=az_accel,
                               reset=reset)
    check_response(acu, resp)



import os
os.environ["OCS_CONFIG_DIR"] = "./test_util/"

from unittest.mock import MagicMock, patch

import pytest

from ocs.ocs_client import OCSReply

from sorunlib import smurf
from util import create_patch_clients

os.environ["SORUNLIB_CONFIG"] = "./data/example_config.yaml"


# Use pytest-mock plugin to patch CLIENTS on all tests
patch_clients = create_patch_clients('satp', autouse=True)


def test_set_targets():
    smurf.set_targets(['smurf1'])
    assert len(smurf.run.CLIENTS['smurf']) == 1
    assert smurf.run.CLIENTS['smurf'][0].instance_id == 'smurf1'


@patch('sorunlib.smurf.time.sleep', MagicMock())
@pytest.mark.parametrize("concurrent", [(True), (False)])
def test_bias_step(concurrent):
    smurf.bias_step(concurrent=concurrent, settling_time=10)
    for client in smurf.run.CLIENTS['smurf']:
        client.take_bias_steps.start.assert_called_with(tag=None)


@patch('sorunlib.smurf.time.sleep', MagicMock())
@pytest.mark.parametrize("concurrent", [(True), (False)])
def test_bias_step_single_failure(concurrent):
    # Create failure on smurf1
    mocked_response = OCSReply(
        0, 'msg', {'success': False, 'op_name': 'bias_step'})
    smurf.run.CLIENTS['smurf'][0].take_bias_steps.wait.side_effect = [mocked_response]

    smurf.bias_step(concurrent=concurrent, settling_time=10)
    for client in smurf.run.CLIENTS['smurf']:
        client.take_bias_steps.start.assert_called_once()


@patch('sorunlib.smurf.time.sleep', MagicMock())
@pytest.mark.parametrize("concurrent", [(True), (False)])
def test_bias_step_failure_threshold(concurrent):
    # Create failure on smurf1 and smurf2
    mocked_response = OCSReply(
        0, 'msg', {'success': False, 'op_name': 'bias_step'})
    smurf.run.CLIENTS['smurf'][0].take_bias_steps.wait.side_effect = [mocked_response]
    smurf.run.CLIENTS['smurf'][1].take_bias_steps.wait.side_effect = [mocked_response]

    with pytest.raises(RuntimeError):
        smurf.bias_step(concurrent=concurrent, settling_time=10)


@patch('sorunlib.smurf.time.sleep', MagicMock())
@pytest.mark.parametrize("concurrent", [(True), (False)])
def test_iv_curve(concurrent):
    smurf.iv_curve(concurrent=concurrent)
    for client in smurf.run.CLIENTS['smurf']:
        client.take_iv.start.assert_called_with(tag=None, kwargs=None)


@patch('sorunlib.smurf.time.sleep', MagicMock())
@pytest.mark.parametrize("concurrent", [(True), (False)])
def test_uxm_setup(concurrent):
    smurf.uxm_setup(concurrent=concurrent)
    for client in smurf.run.CLIENTS['smurf']:
        client.uxm_setup.start.assert_called_once()


@patch('sorunlib.smurf.time.sleep', MagicMock())
@pytest.mark.parametrize("concurrent", [(True), (False)])
def test_uxm_relock(concurrent):
    smurf.uxm_relock(concurrent=concurrent)
    for client in smurf.run.CLIENTS['smurf']:
        client.uxm_relock.start.assert_called_once()


@patch('sorunlib.smurf.time.sleep', MagicMock())
@pytest.mark.parametrize("test_mode", [(True), (False)])
def test_uxm_relock_test_mode(test_mode):
    smurf.uxm_relock(test_mode=test_mode)
    for client in smurf.run.CLIENTS['smurf']:
        if test_mode:
            client.uxm_relock.start.assert_called_with(test_mode=test_mode)
        else:
            client.uxm_relock.start.assert_called_with()


@pytest.mark.parametrize("concurrent", [(True), (False)])
def test_bias_dets(concurrent):
    smurf.bias_dets(concurrent=concurrent)
    for client in smurf.run.CLIENTS['smurf']:
        client.bias_dets.start.assert_called_once()


@pytest.mark.parametrize("concurrent", [(True), (False)])
def test_set_biases(concurrent):
    smurf.set_biases(bias=1, bias_group=None, concurrent=concurrent)
    for client in smurf.run.CLIENTS['smurf']:
        client.set_biases.start.assert_called_with(bias=1, bg=None)


@pytest.mark.parametrize("concurrent", [(True), (False)])
def test_zero_biases(concurrent):
    smurf.zero_biases(bias_group=None, concurrent=concurrent)
    for client in smurf.run.CLIENTS['smurf']:
        client.zero_biases.start.assert_called_with(bg=None)


@pytest.mark.parametrize("concurrent", [(True), (False)])
def test_bgmap(concurrent):
    smurf.take_bgmap(concurrent=concurrent)
    for client in smurf.run.CLIENTS['smurf']:
        client.take_bgmap.start.assert_called_with(tag=None)


@pytest.mark.parametrize("concurrent", [(True), (False)])
def test_take_noise(concurrent):
    smurf.take_noise(concurrent=concurrent)
    for client in smurf.run.CLIENTS['smurf']:
        client.take_noise.start.assert_called_with(tag=None)


@pytest.mark.parametrize("concurrent", [(True), (False)])
def test_shutdown(concurrent):
    smurf.shutdown(concurrent=concurrent)
    for client in smurf.run.CLIENTS['smurf']:
        client.all_off.start.assert_called_once()


@pytest.mark.parametrize("state", [("on"), ("off")])
def test_stream(state):
    smurf.stream(state=state)
    for client in smurf.run.CLIENTS['smurf']:
        if state == "on":
            client.stream.start.assert_called_once()
        else:
            client.stream.stop.assert_called_once()


@pytest.mark.parametrize("state", [("on"), ("off")])
def test_stream_single_failure(state):
    # Create failure on smurf1
    mocked_response = OCSReply(
        0, 'msg', {'success': False, 'op_name': 'stream'})
    # For state == 'on'
    smurf.run.CLIENTS['smurf'][0].stream.status.side_effect = [mocked_response]
    # For state == 'off'
    smurf.run.CLIENTS['smurf'][0].stream.wait.side_effect = [mocked_response]

    smurf.stream(state=state)
    for client in smurf.run.CLIENTS['smurf']:
        if state == "on":
            client.stream.start.assert_called_once()
        else:
            client.stream.stop.assert_called_once()



import os
os.environ["OCS_CONFIG_DIR"] = "./test_util/"

import pytest

from sorunlib import hwp

from util import create_patch_clients


patch_clients_satp = create_patch_clients('satp')


@pytest.mark.parametrize("active", [True, False])
def test_stop(patch_clients_satp, active):
    hwp.stop(active=active)
    if active:
        hwp.run.CLIENTS['hwp'].brake.assert_called_with()
    else:
        hwp.run.CLIENTS['hwp'].pmx_off.assert_called()


def test_stop_brake_voltage(patch_clients_satp):
    VOLTAGE = 5.0
    hwp.stop(active=True, brake_voltage=VOLTAGE)
    hwp.run.CLIENTS['hwp'].brake.assert_called_with(brake_voltage=VOLTAGE)


def test_set_freq(patch_clients_satp):
    hwp.set_freq(freq=2.0)
    hwp.run.CLIENTS['hwp'].pid_to_freq.assert_called_with(target_freq=2.0)



import os
import time
os.environ["OCS_CONFIG_DIR"] = "./test_util/"

import pytest

from unittest.mock import MagicMock, patch

import ocs
from ocs.ocs_client import OCSReply
from sorunlib import wiregrid

from util import create_session, create_patch_clients, mocked_clients

patch_clients = create_patch_clients('satp')


def create_acu_client(az, el, boresight):
    """Create an ACU client with mock monitor Process session.data.

    Args:
        az (float): Azimuth position.
        el (float): Elevation position.
        boresight (float): Boresight position.

    """
    acu_client = MagicMock()
    session = create_session('monitor')
    session.data = {'StatusDetailed': {'Azimuth current position': az,
                                       'Elevation current position': el,
                                       'Boresight current position': boresight}}
    reply = OCSReply(ocs.OK, 'msg', session.encoded())
    acu_client.monitor.status = MagicMock(return_value=reply)

    return acu_client


def create_labjack_client():
    """Create a labjack client with mock acq Process session.data."""
    client = MagicMock()
    session = create_session('acq')
    session.data = {
        "block_name": "sens",
        "data": {
            "AIN0C": 23.0,
            "AIN1C": 20.0,
            "AIN2C": 30.0,
        },
        "timestamp": time.time()
    }
    session.set_status('running')
    reply = OCSReply(ocs.OK, 'msg', session.encoded())
    client.acq.status = MagicMock(return_value=reply)

    return client


def create_actuator_client(motor):
    """Create an actuator client with mock acq Process session.data.

    Args:
        motor (int): Motor state, 0 is off, 1 is on.

    """
    client = MagicMock()
    session = create_session('acq')
    session.data = {'fields': {'motor': motor},
                    'timestamp': time.time()}
    session.set_status('running')
    reply = OCSReply(ocs.OK, 'msg', session.encoded())
    client.acq.status = MagicMock(return_value=reply)

    return client


def create_kikusui_client():
    """Create a kikusui client with mock IV_acq Process session.data."""
    client = MagicMock()
    session = create_session('IV_acq')
    session.data = {
        'fields': {
            'kikusui': {
                'volt': 12.0,
                'curr': 3.0,
                'voltset': 12.0,
                'currset': 3.0,
                'status': 1,
            }
        },
        'timestamp': time.time(),
    }
    session.set_status('running')
    reply = OCSReply(ocs.OK, 'msg', session.encoded())
    client.IV_acq.status = MagicMock(return_value=reply)

    return client


def create_encoder_client():
    """Create an encoder client with mock acq Process session.data."""
    client = MagicMock()
    session = create_session('acq')
    session.data = {
        'fields': {
            'encoder_data': {'last_updated': time.time()},
        },
        'timestamp': time.time(),
    }
    session.set_status('running')
    reply = OCSReply(ocs.OK, 'msg', session.encoded())
    client.acq.status = MagicMock(return_value=reply)

    return client


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
def test_insert():
    wiregrid.insert()
    wiregrid.run.CLIENTS['wiregrid']['actuator'].insert.assert_called_with()


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
def test_insert_failed():
    mocked_response = OCSReply(
        0, 'msg', {'success': False, 'op_name': 'insert'})
    wiregrid.run.CLIENTS['wiregrid']['actuator'].insert.side_effect = [mocked_response]
    with pytest.raises(RuntimeError):
        wiregrid.insert()


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
def test_eject():
    wiregrid.eject()
    wiregrid.run.CLIENTS['wiregrid']['actuator'].eject.assert_called_with()


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
def test_eject_failed():
    mocked_response = OCSReply(
        0, 'msg', {'success': False, 'op_name': 'eject'})
    wiregrid.run.CLIENTS['wiregrid']['actuator'].eject.side_effect = [mocked_response]
    with pytest.raises(RuntimeError):
        wiregrid.eject()


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
@patch('sorunlib.wiregrid.time.sleep', MagicMock())
def test_rotate_continuous():
    wiregrid.rotate(continuous=True)
    wiregrid.run.CLIENTS['wiregrid']['kikusui'].set_v.assert_called_once_with(volt=12)
    wiregrid.run.CLIENTS['wiregrid']['kikusui'].set_c.assert_called_once_with(current=3.0)
    wiregrid.run.CLIENTS['wiregrid']['kikusui'].set_on.assert_called_once()
    wiregrid.run.CLIENTS['wiregrid']['kikusui'].set_off.assert_called_once()


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
@patch('sorunlib.wiregrid.time.sleep', MagicMock())
def test_rotate_stepwise():
    wiregrid.rotate(continuous=False)
    wiregrid.run.CLIENTS['wiregrid']['kikusui'].set_v.assert_called_once_with(volt=12)
    wiregrid.run.CLIENTS['wiregrid']['kikusui'].set_c.assert_called_once_with(current=2.4)
    wiregrid.run.CLIENTS['wiregrid']['kikusui'].stepwise_rotation.assert_called_once_with(num_laps=1, stopped_time=10)


def test__verify_temp_response():
    session = create_session('test')
    session.data = {"data": {"TEST": 2}}
    reply = OCSReply(ocs.OK, 'msg', session.encoded())
    wiregrid._verify_temp_response(reply, 'TEST', 0)


def test__verify_temp_response_invalid():
    session = create_session('test')
    session.data = {"data": {"TEST": -10}}
    reply = OCSReply(ocs.OK, 'msg', session.encoded())
    with pytest.raises(RuntimeError):
        wiregrid._verify_temp_response(reply, 'TEST', 0)


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
def test__check_temperature_sensors():
    wiregrid.run.CLIENTS['wiregrid']['labjack'] = create_labjack_client()
    wiregrid._check_temperature_sensors()
    wiregrid.run.CLIENTS['wiregrid']['labjack'].acq.status.assert_called_once()


def test__check_telescope_position(patch_clients):
    wiregrid._check_telescope_position()
    wiregrid.run.CLIENTS['acu'].monitor.status.assert_called_once()


@pytest.mark.parametrize('el,boresight', [(40, 0), (50, 10)])
@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
def test__check_telescope_position_invalid(el, boresight):
    wiregrid.run.CLIENTS['acu'] = create_acu_client(180, el, boresight)
    with pytest.raises(RuntimeError):
        wiregrid._check_telescope_position()
    wiregrid.run.CLIENTS['acu'].monitor.status.assert_called_once()


@pytest.mark.parametrize('motor', [(0), (1)])
@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
def test__check_motor_on(motor):
    wiregrid.run.CLIENTS['wiregrid']['actuator'] = create_actuator_client(motor=motor)
    wiregrid._check_motor_on()
    if motor == 1:
        wiregrid.run.CLIENTS['wiregrid']['actuator'].acq.status.assert_called_once()
        wiregrid.run.CLIENTS['wiregrid']['actuator'].motor_on.assert_not_called()
    elif motor == 0:
        wiregrid.run.CLIENTS['wiregrid']['actuator'].acq.status.assert_called_once()
        wiregrid.run.CLIENTS['wiregrid']['actuator'].motor_on.assert_called_once()


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
def test__check_motor_on_invalid_state():
    wiregrid.run.CLIENTS['wiregrid']['actuator'] = create_actuator_client(motor=3)
    with pytest.raises(RuntimeError):
        wiregrid._check_motor_on()
    wiregrid.run.CLIENTS['wiregrid']['actuator'].acq.status.assert_called_once()


@patch('sorunlib.wiregrid.run.CLIENTS', mocked_clients())
def test__check_agents_online():
    wiregrid.run.CLIENTS['wiregrid']['actuator'] = create_actuator_client(motor=1)
    wiregrid.run.CLIENTS['wiregrid']['kikusui'] = create_kikusui_client()
    wiregrid.run.CLIENTS['wiregrid']['encoder'] = create_encoder_client()
    wiregrid.run.CLIENTS['wiregrid']['labjack'] = create_labjack_client()
    wiregrid._check_agents_online()
    wiregrid.run.CLIENTS['wiregrid']['actuator'].acq.status.assert_called_once()
    wiregrid.run.CLIENTS['wiregrid']['kikusui'].IV_acq.status.assert_called_once()
    wiregrid.run.CLIENTS['wiregrid']['encoder'].acq.status.assert_called()
    wiregrid.run.CLIENTS['wiregrid']['labjack'].acq.status.assert_called_once()


@pytest.mark.parametrize('continuous', [(True), (False)])
@patch('sorunlib.wiregrid.time.sleep', MagicMock())
def test_calibrate_stepwise(patch_clients, continuous):
    # Setup all mock clients
    wiregrid.run.CLIENTS['acu'] = create_acu_client(180, 50, 0)
    wiregrid.run.CLIENTS['wiregrid']['actuator'] = create_actuator_client(motor=1)
    wiregrid.run.CLIENTS['wiregrid']['kikusui'] = create_kikusui_client()
    wiregrid.run.CLIENTS['wiregrid']['encoder'] = create_encoder_client()
    wiregrid.run.CLIENTS['wiregrid']['labjack'] = create_labjack_client()

    wiregrid.calibrate(continuous=continuous)
    # All other internal functions tested separately, just make sure smurf
    # stream is run
    for client in wiregrid.run.CLIENTS['smurf']:
        client.stream.start.assert_called()
        client.stream.stop.assert_called()


def test__check_process_data_stale_data():
    with pytest.raises(RuntimeError):
        stale_time = time.time() - wiregrid.AGENT_TIMEDIFF_THRESHOLD
        wiregrid._check_process_data('test process', stale_time)



from setuptools import setup, find_packages

import versioneer

with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='sorunlib',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='OCS Control Programs for running the observatory.',
    long_description=long_description,
    long_description_content_type="text/x-rst",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    url="https://github.com/simonsobs/sorunlib",
    project_urls={
        "Source Code": "https://github.com/simonsobs/sorunlib",
        "Documentation": "https://sorunlib.readthedocs.io",
        "Bug Tracker": "https://github.com/simonsobs/sorunlib/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: BSD License",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Astronomy",
    ],
    python_requires=">=3.7",
    install_requires=[
        'ocs==0.11.0',
        'pyyaml',
    ],
    extras_require={
        "tests": ["pytest>=7.0.0", "pytest-cov>=3.0.0", "pytest-mock>=3.12.0"],
        "docs": ["sphinx>=5.3.0", "sphinx_rtd_theme>=1.1.1"],
    },
)



"""This module is responsible for commanding all active pysmurf-controller
agents.

If a SMuRF fails to perform an operation it will be dropped from the global
``CLIENTS`` list, removing it from further use during that schedule. If the
number of active SMuRFs falls below the configured ``smurf_failure_threshold``
an exception will be raised, halting observations.

:func:`initialize() <sorunlib.__init__.initialize>` must be run to re-add
previously failed SMuRFs to the ``CLIENTS`` list.

"""

import time

import sorunlib as run
from sorunlib._internal import check_response, check_started

# Timing between commanding separate SMuRF Controllers
# Yet to be determined in the field. Eventually might need this to be unique
# per operation. Also, move to configuration file once sorunlib has one.
CRYO_WAIT = 120


def _wait_for_cryo(time_):
    if time_ is None:
        wait = CRYO_WAIT
    else:
        wait = time_
    time.sleep(wait)


def _check_smurf_threshold():
    cfg = run.config.load_config()
    threshold = cfg['smurf_failure_threshold']
    remaining = len(run.CLIENTS['smurf'])
    if remaining < threshold:
        error = 'Functional SMuRF count below failure threshold ' + \
                f'({remaining} < {threshold}). Aborting.'
        raise RuntimeError(error)


def _run_op(operation, concurrent, settling_time, **kwargs):
    """Run operation across all active SMuRF controllers.

    Additional kwargs are passed directly to the ``.start()`` call.

    Args:
        operation (str): Operation name.
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controllers if *not* running concurrently. If running
            concurrently this is ignored. If None, defaults to a fixed wait
            time of 120 seconds.

    """
    clients_to_remove = []

    # Start operation
    for smurf in run.CLIENTS['smurf']:
        op = smurf.__getattribute__(operation)
        op.start(**kwargs)
        # Run in series
        if not concurrent:
            resp = op.wait()
            try:
                check_response(smurf, resp)
            except RuntimeError as e:
                print(f"Failed to perform {operation} on {smurf}, removing from targets list.")
                print(e)
                clients_to_remove.append(smurf)

            # Allow cryo to settle
            _wait_for_cryo(settling_time)

    # Run in parallel
    if concurrent:
        for smurf in run.CLIENTS['smurf']:
            op = smurf.__getattribute__(operation)
            resp = op.wait()
            try:
                check_response(smurf, resp)
            except RuntimeError as e:
                print(f"Failed to perform {operation} on {smurf}, removing from targets list.")
                print(e)
                clients_to_remove.append(smurf)

    # Remove failed SMuRF clients
    for client in clients_to_remove:
        run.CLIENTS['smurf'].remove(client)

    # Check if enough SMuRFs remain
    _check_smurf_threshold()


def set_targets(targets):
    """Set the target pysmurf-controller Agents that sorunlib will command.

    Args:
        targets (list): List of pysmurf-controller instance-ids to send future
            commands to. This must be a subset of the currently active clients
            list.

    Notes:
        This modifies the global ``sorunlib.CLIENTS`` list.

    """
    _smurf_clients = []
    for smurf in run.CLIENTS['smurf']:
        if smurf.instance_id in targets:
            _smurf_clients.append(smurf)

    run.CLIENTS['smurf'] = _smurf_clients


def bias_step(tag=None, concurrent=True, settling_time=None):
    """Perform a bias step on all SMuRF Controllers.

    Args:
        tag (str, optional): Tag or comma-separated listed of tags to attach to
            the operation.
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controllers if *not* running concurrently. If running
            concurrently this is ignored. If None, defaults to a fixed wait
            time of 120 seconds.

    """
    _run_op('take_bias_steps',
            concurrent=concurrent,
            settling_time=settling_time,
            tag=tag)


def iv_curve(tag=None, iv_kwargs=None, concurrent=True, settling_time=None):
    """Perform an iv curve on all SMuRF Controllers.

    Args:
        tag (str, optional): Tag or comma-separated listed of tags to attach to
            the operation.
        iv_kwargs (dict, optional): Additional keyword arguments to pass to
            ``take_iv``.
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controllers if *not* running concurrently. If running
            concurrently this is ignored. If None, defaults to a fixed wait
            time of 120 seconds.

    """
    _run_op('take_iv',
            concurrent=concurrent,
            settling_time=settling_time,
            tag=tag,
            kwargs=iv_kwargs)


def uxm_setup(concurrent=True, settling_time=0):
    """Perform first-time setup procedure for a UXM.

    Args:
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controllers if *not* running concurrently. If running
            concurrently this is ignored. Defaults to 0 seconds.

    """
    _run_op('uxm_setup',
            concurrent=concurrent,
            settling_time=settling_time)


def uxm_relock(test_mode=False, concurrent=True, settling_time=0):
    """Relocks detectors to existing tune if setup has already been run.

    Args:
        test_mode (bool): Run uxm_setup() task in test_mode, removing emulated
            wait times.
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controllers if *not* running concurrently. If running
            concurrently this is ignored. Defaults to 0 seconds.

    """
    if test_mode:
        _run_op('uxm_relock',
                concurrent=concurrent,
                settling_time=settling_time,
                test_mode=test_mode)
    else:
        _run_op('uxm_relock',
                concurrent=concurrent,
                settling_time=settling_time)


def bias_dets(concurrent=True, settling_time=0):
    """Bias the detectors on all SMuRF Controllers.

    Args:
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controllers if *not* running concurrently. If running
            concurrently this is ignored. Defaults to 0 seconds.

    """
    _run_op('bias_dets',
            concurrent=concurrent,
            settling_time=settling_time)


def set_biases(bias, bias_group=None, concurrent=True, settling_time=0):
    """Set the detector biases on all SMuRF Controllers.

    Args:
        bias: (int, float, list)
            Biases to set. If a float is passed, this will be used for all
            specified bias groups. If a list of floats is passed, it must be
            the same size of the list of bias groups.
        bias_group: (int, list, optional)
            Bias group or list of bias groups to set. If None, will set all
            bias groups. Defaults to None.
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controlls if *not* running concurrently. If running
            concurrently this is ignored. Defaults to 0 seconds.

    """
    _run_op('set_biases',
            concurrent=concurrent,
            settling_time=settling_time,
            bias=bias,
            bg=bias_group)


def zero_biases(bias_group=None, concurrent=True, settling_time=0):
    """Set the detector biases on all SMuRF Controllers.

    Args:
        bias_group: (int, list, optional)
            Bias group or list of bias groups to set. If None, will zero all
            bias groups. Defaults to None.
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controlls if *not* running concurrently. If running
            concurrently this is ignored. Defaults to 0 seconds.

    """
    _run_op('zero_biases',
            concurrent=concurrent,
            settling_time=settling_time,
            bg=bias_group)


def take_bgmap(tag=None, concurrent=True, settling_time=0):
    """Take a bgmap on all SMuRF Controllers.

    Args:
        tag (str, optional): Tag or comma-separated listed of tags to attach to
            the operation.
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controllers if *not* running concurrently. If running
            concurrently this is ignored. Defaults to 0 seconds.

    """
    _run_op('take_bgmap',
            concurrent=concurrent,
            settling_time=settling_time,
            tag=tag)


def take_noise(tag=None, concurrent=True, settling_time=0):
    """Measure noise statistics from a short, 30 second, timestream.

    Args:
        tag (str, optional): Tag or comma-separated listed of tags to attach to
            the operation.
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controllers if *not* running concurrently. If running
            concurrently this is ignored. Defaults to 0 seconds.

    """
    _run_op('take_noise',
            concurrent=concurrent,
            settling_time=settling_time,
            tag=tag)


def shutdown(concurrent=True, settling_time=0):
    """Safely shutdown the SMuRFs.

    This turns off tones, flux-ramp voltage and amplifier biases on all SMuRF
    Controllers.

    Args:
        concurrent (bool, optional): A bool which determines how the operation
            is run across the active SMuRF controllers. It runs in parallel if
            True, and in series if False.
        settling_time (float, optional):
            Time in seconds to wait between operation runs across the active
            SMuRF controllers if *not* running concurrently. If running
            concurrently this is ignored. Defaults to 0 seconds.

    """
    _run_op('all_off',
            concurrent=concurrent,
            settling_time=settling_time)


def stream(state, tag=None, subtype=None):
    """Stream data on all SMuRF Controllers.

    Args:
        state (str): Streaming state, either 'on' or 'off'.
        tag (str, optional): Tag or comma-separated listed of tags to attach to
            the operation.
        subtype (str, optional): Operation subtype used to tag the stream.

    """
    clients_to_remove = []

    if state.lower() == 'on':
        for smurf in run.CLIENTS['smurf']:
            smurf.stream.start(subtype=subtype, tag=tag)

        for smurf in run.CLIENTS['smurf']:
            resp = smurf.stream.status()
            try:
                check_started(smurf, resp, timeout=120)
            except RuntimeError as e:
                print(f"Failed to start stream on {smurf}, removing from targets list.")
                print(e)
                smurf.stream.stop()
                clients_to_remove.append(smurf)

    else:
        for smurf in run.CLIENTS['smurf']:
            print(f'Turning off stream from {smurf.instance_id}.')
            smurf.stream.stop()
            resp = smurf.stream.wait()
            try:
                check_response(smurf, resp)
            except RuntimeError as e:
                print(f"Failed to stop stream on {smurf}, removing from targets list.")
                print(e)
                clients_to_remove.append(smurf)

    # Remove failed SMuRF clients
    for client in clients_to_remove:
        run.CLIENTS['smurf'].remove(client)

    # Check if enough SMuRFs remain
    _check_smurf_threshold()
